import click
import os
import pendulum
from click.exceptions import Abort
from tabulate import tabulate

from prefect import Client, config
from prefect.exceptions import AuthorizationError, ClientError
from prefect.cli.build_register import handle_terminal_error, TerminalError
from prefect.backend import TenantView


def check_override_auth_token():
    if config.cloud.get("auth_token"):
        if os.environ.get("PREFECT__CLOUD__AUTH_TOKEN"):
            click.secho(
                "Auth token has been set in the environment. The CLI cannot be used to "
                "manage your auth. Unset the key with `unset PREFECT__CLOUD__AUTH_TOKEN`",
                fg="red",
            )
        else:
            click.secho(
                "Auth token has been set in the config. The CLI cannot be used to "
                "manage your auth. Remove the key from your configuration file.",
                fg="red",
            )
        raise Abort


def abort_on_config_api_key(message: str = None):
    if config.cloud.get("api_key"):
        # Add a leading space if not null
        message = (" " + message) if message else ""
        raise TerminalError(
            "Your API key is set in the Prefect config instead of with the CLI."
            + message
        )


@click.group(hidden=True)
def auth():
    """
    Handle Prefect Cloud authorization.

    \b
    Usage:
        $ prefect auth [COMMAND]

    \b
    Commands:
        login           Log in to Prefect Cloud
        logout          Log out of Prefect Cloud
        list-tenants    List your available tenants
        switch-tenants  Switch to a different tenant
        create-key      Create an API key
        list-keys       List details of existing API keys
        revoke-key      Delete an API key from the backend
        create-token    Create an API token (DEPRECATED)
        list-tokens     List the names and ids of existing API tokens (DEPRECATED)
        revoke-token    Delete an API token from the backend (DEPRECATED)

    \bExamples:

    \b  Log in using an existing key

    \b    $ prefect auth login --key MY_KEY

    \b  Log out, removing your current key or token

    \b    $ prefect auth logout

    \b  Switch to another tenant by slug

    \b    $ prefect auth switch-tenants --slug test-person

    \b  Create a new API key that expires at the start of the next year

    \b    $ prefect auth create-key -n marvin --expire 2022-1-1
    """
    if config.backend == "server":
        raise click.UsageError(
            "Prefect Server does not have authentication. Change your backend to "
            "Prefect Cloud with `prefect backend cloud` to log in."
        )


@auth.command(hidden=True)
@click.option(
    "--key",
    "-k",
    help="A Prefect Cloud API key.",
)
@click.option(
    "--token",
    "-t",
    help="A Prefect Cloud API token. DEPRECATED.",
)
@handle_terminal_error
def login(key, token):
    """
    Login to Prefect Cloud

    Create an API key in the UI then login with it here:

        $ prefect auth login -k YOUR-KEY

    You will be switched to the default tenant associated with the key. After login,
    your available tenants can be seen with `prefect auth list-tenants` and you can
    change the default tenant on this machine using `prefect auth switch-tenants`.

    The given key will be stored on disk for later access. Prefect will default to using
    this key for all interaction with the API but frequently overrides can be passed to
    individual commands or functions. To remove your key from disk, see
    `prefect auth logout`.

    This command has backwards compatibility support for API tokens, which are a
    deprecated form of authentication with Prefect Cloud
    """
    if not key and not token:
        raise TerminalError("You must supply an API key or token!")

    if key and token:
        raise TerminalError("You cannot supply both an API key and token")

    abort_on_config_api_key(
        "To log in with the CLI, remove the config key `prefect.cloud.api_key`"
    )

    # Attempt to treat the input like an API key even if it is passed as a token
    # Ignore any tenant id that has been previously set via login
    client = Client(api_key=key or token, tenant_id=None)

    try:
        tenant_id = client._get_auth_tenant()
    except AuthorizationError:
        if key:  # We'll catch an error again later if using a token
            raise TerminalError("Unauthorized. Invalid Prefect Cloud API key.")
    except ClientError:
        raise TerminalError("Error attempting to communicate with Prefect Cloud.")
    else:
        if token:
            click.secho(
                "WARNING: You logged in with an API key using the `--token` flag "
                "which is deprecated. Please use `--key` instead.",
                fg="yellow",
            )
        client.tenant_id = tenant_id
        client.save_auth_to_disk()
        tenant = TenantView.from_tenant_id(tenant_id)
        click.secho(
            f"Logged in to Prefect Cloud tenant {tenant.name!r} ({tenant.slug})",
            fg="green",
        )
        return

        # If there's not a tenant id, we've been given an actual token, fallthrough to
        # the backwards compatibility token auth

    # Backwards compatibility for tokens
    if token:
        check_override_auth_token()
        client = Client(api_token=token)

        # Verify they're not also using an API key
        if client.api_key:
            raise TerminalError(
                "You have already logged in with an API key and cannot use a token."
            )

        click.secho(
            "WARNING: API tokens are deprecated. Please create an API key and use "
            "`prefect auth login --key <KEY>` to login instead.",
            fg="yellow",
        )

        # Verify login obtained a valid api token
        try:
            output = client.graphql(
                query={"query": {"user": {"default_membership": "tenant_id"}}}
            )

            # Log into default membership
            success_login = client.login_to_tenant(
                tenant_id=output.data.user[0].default_membership.tenant_id
            )

            if not success_login:
                raise AuthorizationError

        except AuthorizationError:
            click.secho(
                "Error attempting to use the given API token. "
                "Please check that you are providing a USER scoped Personal Access Token "
                "and consider switching API key.",
                fg="red",
            )
            return
        except ClientError:
            click.secho(
                "Error attempting to communicate with Prefect Cloud. "
                "Please check that you are providing a USER scoped Personal Access Token "
                "and consider switching API key.",
                fg="red",
            )
            return

        # save token
        client.save_api_token()

        click.secho("Login successful!", fg="green")


@auth.command(hidden=True)
@click.option(
    "--token",
    "-t",
    help="Log out from the API token based authentication, ignoring API keys",
    is_flag=True,
)
@handle_terminal_error
def logout(token):
    """
    Log out of Prefect Cloud

    This will remove your cached authentication from disk.
    """

    client = Client()

    # Log out of API keys unless given the token flag
    if client.api_key and not token:

        # Check the source of the API key
        abort_on_config_api_key(
            "To log out, remove the config key `prefect.cloud.api_key`"
        )

        click.confirm(
            "Are you sure you want to log out of Prefect Cloud? "
            "This will remove your API key from this machine.",
            default=False,
            abort=True,
        )

        # Clear the key and tenant id then write to the cache
        client.api_key = ""
        client._tenant_id = ""
        client.save_auth_to_disk()

        click.secho("Logged out of Prefect Cloud", fg="green")

    elif client._api_token:

        check_override_auth_token()
        tenant_id = client.active_tenant_id

        if not tenant_id:
            click.confirm(
                "Are you sure you want to log out of Prefect Cloud? "
                "This will remove your API token from this machine.",
                default=False,
                abort=True,
            )

            # Remove the token from local storage by writing blank settings
            client._save_local_settings({})
            click.secho("Logged out of Prefect Cloud", fg="green")

        else:
            # Log out of the current tenant (dropping the access token) while retaining
            # the API token. This is backwards compatible behavior. Running the logout
            # command twice will remove the token from storage entirely
            click.confirm(
                "Are you sure you want to log out of your current Prefect Cloud tenant?",
                default=False,
                abort=True,
            )

            client.logout_from_tenant()

            click.secho(
                f"Logged out from tenant {tenant_id}. Run `prefect auth logout` again "
                "to delete your API token.",
                fg="green",
            )
    else:
        raise TerminalError(
            "You are not logged in to Prefect Cloud. "
            "Use `prefect auth login` to log in first."
        )


@auth.command(hidden=True)
@handle_terminal_error
def list_tenants():
    """
    List available tenants
    """
    client = Client()

    try:
        tenants = client.get_available_tenants()
    except AuthorizationError:
        raise TerminalError(
            "You are not authenticated. Use `prefect auth login` first."
        )

    output = []
    for item in tenants:
        active = None
        if item.id == client.tenant_id:
            active = "*"
        output.append([item.name, item.slug, item.id, active])

    click.echo(
        tabulate(
            output,
            headers=["NAME", "SLUG", "ID", ""],
            tablefmt="plain",
            numalign="left",
            stralign="left",
        )
    )


@auth.command(hidden=True)
@click.option(
    "--id", "-i", required=False, help="A Prefect Cloud tenant id.", hidden=True
)
@click.option(
    "--slug", "-s", required=False, help="A Prefect Cloud tenant slug.", hidden=True
)
@click.option(
    "--default",
    "-d",
    is_flag=True,
    help="Switch to the default tenant for your API key",
)
@handle_terminal_error
def switch_tenants(id, slug, default):
    """
    Switch active tenant

    \b
    Options:
        --id, -i    TEXT    A Prefect Cloud tenant id
        --slug, -s  TEXT    A Prefect Cloud tenant slug
    """

    # If the config specifies a tenant explicitly, it is used before this mechanism
    if config.cloud.get("tenant_id"):
        raise TerminalError(
            "Your tenant id has been set in the Prefect config instead of with the "
            "CLI. To switch tenants with the CLI, remove the config key "
            " `prefect.cloud.tenant_id`"
        )

    client = Client()

    # Deprecated API token check
    if not client.api_key:
        check_override_auth_token()

        if default:
            raise TerminalError(
                "The default tenant flag can only be used with API keys."
            )

    else:  # Using an API key
        if default:
            # Clear the set tenant on disk
            client.tenant_id = None
            client.save_auth_to_disk()
            click.secho(
                "Tenant restored to the default tenant for your API key: "
                f"{client._get_auth_tenant()}",
                fg="green",
            )
            return

    login_success = client.login_to_tenant(tenant_slug=slug, tenant_id=id)
    if not login_success:
        raise TerminalError("Unable to switch tenant!")

    # `login_to_tenant` will write to disk if using an API token, if using an API key
    # we will write to disk manually here
    if client.api_key:
        client.save_auth_to_disk()

    click.secho(f"Tenant switched to {client.tenant_id}", fg="green")


@auth.command(hidden=True)
@click.option("--name", "-n", required=True, help="A token name.", hidden=True)
@click.option("--scope", "-s", required=True, help="A token scopre.", hidden=True)
def create_token(name, scope):
    """
    DEPRECATED. Please use API keys instead.

    Create a Prefect Cloud API token.

    For more info on API tokens visit https://docs.prefect.io/orchestration/concepts/api.html

    \b
    Options:
        --name, -n      TEXT    A name to give the generated token
        --scope, -s     TEXT    A scope for the token
    """
    click.secho(
        "WARNING: API tokens are deprecated. Please use `prefect auth create-key` to "
        "create an API key instead.",
        fg="yellow",
        err=True,  # Write to stderr in case the user is piping
    )

    client = Client()

    output = client.graphql(
        query={
            "mutation($input: create_api_token_input!)": {
                "create_api_token(input: $input)": {"token"}
            }
        },
        variables=dict(input=dict(name=name, scope=scope)),
    )

    if not output.get("data", None):
        click.secho("Issue creating API token", fg="red")
        return

    click.echo(output.data.create_api_token.token)


@auth.command(hidden=True)
def list_tokens():
    """
    DEPRECATED. Please use API keys instead.

    List your available Prefect Cloud API tokens.
    """
    click.secho(
        "WARNING: API tokens are deprecated. Please consider removing your remaining "
        "tokens and using API keys instead.",
        fg="yellow",
        err=True,  # Write to stderr in case the user is piping
    )

    client = Client()
    output = client.graphql(query={"query": {"api_token": {"id", "name"}}})

    if not output.get("data", None):
        click.secho("Unable to list API tokens", fg="red")
        return

    tokens = []
    for item in output.data.api_token:
        tokens.append([item.name, item.id])

    click.echo(
        tabulate(
            tokens,
            headers=["NAME", "ID"],
            tablefmt="plain",
            numalign="left",
            stralign="left",
        )
    )


@auth.command(hidden=True)
@click.option("--id", "-i", required=True, help="A token ID.", hidden=True)
def revoke_token(id):
    """
    DEPRECATED. Please use API keys instead.

    Revote a Prefect Cloud API token

    \b
    Options:
        --id, -i    TEXT    The id of a token to revoke
    """
    check_override_auth_token()

    client = Client()

    output = client.graphql(
        query={
            "mutation($input: delete_api_token_input!)": {
                "delete_api_token(input: $input)": {"success"}
            }
        },
        variables=dict(input=dict(token_id=id)),
    )

    if not output.get("data", None) or not output.data.delete_api_token.success:
        click.secho("Unable to revoke token with ID {}".format(id), fg="red")
        return

    click.secho("Token successfully revoked", fg="green")


@auth.command(hidden=True)
@click.option("--name", "-n", required=True, help="A name to associate with the key")
@click.option(
    "--expire",
    "-e",
    help=(
        "A optional dateutil parsable time to at. "
        "If not given, the key will never expire."
    ),
    default=None,
)
@click.option(
    "--quiet",
    "-q",
    help="If set, only display the created key.",
    is_flag=True,
)
@handle_terminal_error
def create_key(name, expire, quiet):
    """
    Create a Prefect Cloud API key for authentication with your current user
    """
    # TODO: Add service account associated key creation eventually

    # Parse the input expiration
    if expire is not None:
        try:
            expires_at = pendulum.parse(expire, strict=False)
        except pendulum.parsing.exceptions.ParserError as exc:
            raise TerminalError(
                f"Failed to parse expiration time. {exc}\n"
                "Please pass a date in a dateutil parsable format."
            )

        if expires_at.diff(abs=False).in_seconds() > 0:
            raise TerminalError(
                f"Given expiration time {expire!r} is a time in the past: {expires_at}"
            )
        expire_msg = f" that will expire {expires_at.diff_for_humans()}"
    else:
        expires_at = None
        expire_msg = ""

    client = Client()

    # We must retrieve our own user id first since you could be creating a key for a SA
    if not quiet:
        click.echo("Retrieving user information...")

    response = client.graphql({"query": {"auth_info": {"user_id"}}})
    user_id = response.get("data", {}).get("auth_info", {}).get("user_id")
    if not user_id:
        raise TerminalError("Failed to retrieve the current user id from Prefect Cloud")

    # Actually create the key
    if not quiet:
        click.echo(f"Creating key{expire_msg}...")
    response = client.graphql(
        query={
            "mutation($input: create_api_key_input!)": {
                "create_api_key(input: $input)": {"key"}
            }
        },
        variables=dict(
            input=dict(
                name=name,
                user_id=user_id,
                expires_at=expires_at.in_tz("utc").isoformat() if expires_at else None,
            )
        ),
    )

    key = response.get("data", {}).get("create_api_key", {}).get("key")
    if key is None:
        raise TerminalError(f"Unexpected response from Prefect Cloud: {response}")

    if quiet:
        click.echo(key)
    else:
        click.echo(
            "This is the only time this key will be displayed! Store it somewhere safe."
        )
        click.secho(f"Successfully created key: {key}", fg="green")


@auth.command(hidden=True)
@handle_terminal_error
def list_keys():
    """
    List Prefect Cloud API keys

    If you are a tenant admin, this should list all service account keys as well as keys
    you have created.
    """
    client = Client()

    response = client.graphql(
        query={
            "query": {
                "auth_api_key": {
                    "id": True,
                    "name": True,
                    "expires_at": True,
                }
            }
        }
    )
    keys = response.get("data", {}).get("auth_api_key")
    if keys is None:
        raise TerminalError(f"Unexpected response from Prefect Cloud: {response}")

    if not keys:
        click.secho("You have not created any API keys", fg="yellow")

    else:
        click.echo(
            tabulate(
                [(key.name, key.id, key.expires_at or "NEVER") for key in keys],
                headers=["NAME", "ID", "EXPIRES AT"],
                tablefmt="plain",
                numalign="left",
                stralign="left",
            )
        )


@auth.command(hidden=True)
@click.option(
    "--id",
    "-i",
    required=True,
    help="The UUID for the API key to delete.",
)
@handle_terminal_error
def revoke_key(id):
    """
    Revoke a Prefect Cloud API key.
    """
    client = Client()

    output = client.graphql(
        query={
            "mutation($input: delete_api_key_input!)": {
                "delete_api_key(input: $input)": {"success"}
            }
        },
        variables=dict(input=dict(key_id=id)),
    )

    if not output.get("data", None) or not output.data.delete_api_key.success:
        raise TerminalError(f"Unable to revoke key {id!r}")

    click.secho("Key successfully revoked!", fg="green")


@auth.command(hidden=True)
@handle_terminal_error
def status():
    """
    Get the current Prefect authentication status
    """
    client = Client()
    click.echo(f"You are connecting to {client.api_server}")

    if client.api_key:
        click.echo("You are authenticating with an API key")

        try:
            click.echo(f"You are logged in to tenant {client.tenant_id}")
        except Exception as exc:
            click.echo(f"Your authentication is not working: {exc}")

    if client._api_token:
        click.secho(
            "You are logged in with an API token. These have been deprecated in favor "
            "of API keys."
            + (
                " Since you have set an API key as well, this will be ignored."
                if client.api_key
                else ""
            ),
            fg="yellow",
        )

    if not client._api_token and not client.api_key:
        click.secho("You are not logged in!", fg="yellow")



import os
from typing import Any

import yaml

from prefect.core.task import Task
from prefect.tasks.shell import ShellTask
from prefect.utilities.tasks import defaults_from_attrs
from prefect.backend.artifacts import create_markdown_artifact

from .dbt_cloud_utils import (
    DbtCloudListArtifactsFailed,
    list_run_artifact_links,
    trigger_job_run,
    wait_for_job_run,
)


class DbtShellTask(ShellTask):
    """
    Task for running dbt commands. It will create a profiles.yml file prior to running dbt commands.

    This task inherits all configuration options from the
    [ShellTask](https://docs.prefect.io/api/latest/tasks/shell.html#shelltask).

    Args:
        - command (string, optional): dbt command to be executed; can also be
            provided post-initialization by calling this task instance
        - dbt_kwargs (dict, optional): keyword arguments used to populate the profiles.yml file
            (e.g.  `{'type': 'snowflake', 'threads': 4, 'account': '...'}`); can also be
            provided at runtime
        - env (dict, optional): dictionary of environment variables to use for
            the subprocess; can also be provided at runtime
        - environment (string, optional): The default target your dbt project will use
        - overwrite_profiles (boolean, optional): flag to indicate whether existing
            profiles.yml file should be overwritten; defaults to `False`
        - profile_name (string, optional): Profile name used for populating the profile name of
            profiles.yml
        - profiles_dir (string, optional): path to directory where the profile.yml file will be
            contained
        - set_profiles_envar (boolean, optional): flag to indicate whether DBT_PROFILES_DIR
            should be set to the provided profiles_dir; defaults to `True`
        - helper_script (str, optional): a string representing a shell script, which
            will be executed prior to the `command` in the same process. Can be used to
            change directories, define helper functions, etc. when re-using this Task
            for different commands in a Flow; can also be provided at runtime
        - shell (string, optional): shell to run the command with; defaults to "bash"
        - return_all (bool, optional): boolean specifying whether this task should return all
            lines of stdout as a list, or just the last line as a string; defaults to `False`
        - log_stderr (bool, optional): boolean specifying whether this task
            should log the output from stderr in the case of a non-zero exit code;
            defaults to `False`
        - **kwargs: additional keyword arguments to pass to the Task constructor

    Example:
        ```python
        from prefect import Flow
        from prefect.tasks.dbt import DbtShellTask

        with Flow(name="dbt_flow") as f:
            task = DbtShellTask(
                profile_name='default',
                environment='test',
                dbt_kwargs={
                    'type': 'snowflake',
                    'threads': 1,
                    'account': 'account.us-east-1'
                },
                overwrite_profiles=True,
                profiles_dir=test_path
            )(command='dbt run')

        out = f.run()
        ```
    """

    def __init__(
        self,
        command: str = None,
        profile_name: str = None,
        env: dict = None,
        environment: str = None,
        overwrite_profiles: bool = False,
        profiles_dir: str = None,
        set_profiles_envar: bool = True,
        dbt_kwargs: dict = None,
        helper_script: str = None,
        shell: str = "bash",
        return_all: bool = False,
        log_stderr: bool = False,
        **kwargs: Any,
    ):
        self.command = command
        self.profile_name = profile_name
        self.environment = environment
        self.overwrite_profiles = overwrite_profiles
        self.profiles_dir = profiles_dir
        self.set_profiles_envar = set_profiles_envar
        self.dbt_kwargs = dbt_kwargs or {}
        super().__init__(
            **kwargs,
            command=command,
            env=env,
            helper_script=helper_script,
            shell=shell,
            return_all=return_all,
            log_stderr=log_stderr,
        )

    @defaults_from_attrs("command", "env", "helper_script", "dbt_kwargs")
    def run(
        self,
        command: str = None,
        env: dict = None,
        helper_script: str = None,
        dbt_kwargs: dict = None,
    ) -> str:
        """
        If no profiles.yml file is found or if overwrite_profiles flag is set to True, this
        will first generate a profiles.yml file in the profiles_dir directory. Then run the dbt
        cli shell command.

        Args:
            - command (string): shell command to be executed; can also be
                provided at task initialization. Any variables / functions defined in
                `self.helper_script` will be available in the same process this command
                runs in
            - env (dict, optional): dictionary of environment variables to use for
                the subprocess
            - helper_script (str, optional): a string representing a shell script, which
                will be executed prior to the `command` in the same process. Can be used to
                change directories, define helper functions, etc. when re-using this Task
                for different commands in a Flow
            - dbt_kwargs(dict, optional): keyword arguments used to populate the profiles.yml file

        Returns:
            - stdout (string): if `return_all` is `False` (the default), only the last line of
                stdout is returned, otherwise all lines are returned, which is useful for
                passing result of shell command to other downstream tasks. If there is no
                output, `None` is returned.

        Raises:
            - prefect.engine.signals.FAIL: if command has an exit code other
                than 0
        """
        DEFAULT_PROFILES_DIR = os.path.join(os.path.expanduser("~"), ".dbt")
        profiles_exists = False
        if os.getenv("DBT_PROFILES_DIR"):
            dbt_profiles_dir = os.path.expanduser(
                os.getenv("DBT_PROFILES_DIR", DEFAULT_PROFILES_DIR)
            )
            profiles_exists = os.path.exists(
                os.path.join(dbt_profiles_dir, "profiles.yml")
            )
        elif self.profiles_dir:
            profiles_exists = os.path.exists(
                os.path.join(self.profiles_dir, "profiles.yml")
            )
        else:
            profiles_exists = os.path.exists(
                os.path.join(DEFAULT_PROFILES_DIR, "profiles.yml")
            )

        dbt_kwargs = {**self.dbt_kwargs, **(dbt_kwargs or {})}

        if self.overwrite_profiles or not profiles_exists:
            profile = {
                self.profile_name: {
                    "outputs": {self.environment: dbt_kwargs},
                    "target": self.environment,
                }
            }

            if not self.profiles_dir:
                try:
                    os.mkdir(DEFAULT_PROFILES_DIR)
                except OSError:
                    self.logger.warning(
                        "Creation of directory %s has failed" % DEFAULT_PROFILES_DIR
                    )
                profile_path = os.path.join(DEFAULT_PROFILES_DIR, "profiles.yml")
                self.profiles_dir = DEFAULT_PROFILES_DIR
            else:
                profile_path = os.path.join(self.profiles_dir, "profiles.yml")

            with open(profile_path, "w+") as yaml_file:
                yaml.dump(profile, yaml_file, default_flow_style=False)

        if self.set_profiles_envar:
            os.environ["DBT_PROFILES_DIR"] = self.profiles_dir

        return super(DbtShellTask, self).run(
            command=command, env=env, helper_script=helper_script
        )


class DbtCloudRunJob(Task):
    """
    Task for running a dbt Cloud job using dbt Cloud APIs v2.
    For info about dbt Cloud APIs, please refer to https://docs.getdbt.com/dbt-cloud/api-v2
    Please note that this task will fail if any call to dbt Cloud APIs fails.

    Running this task will generate a markdown artifact viewable in the Prefect UI. The artifact
    will contain links to the dbt artifacts generate as a result of the job run.

    Args:
        - cause (string): A string describing the reason for triggering the job run
        - account_id (int, optional): dbt Cloud account ID.
            Can also be passed as an env var.
        - job_id (int, optional): dbt Cloud job ID
        - token (string, optional): dbt Cloud token.
            Please note that this token must have access at least to the dbt Trigger Job API.
        - additional_args (dict, optional): additional information to pass to the Trigger Job API.
            For a list of the possible information,
            have a look at: https://docs.getdbt.com/dbt-cloud/api-v2#operation/triggerRun
        - account_id_env_var_name (string, optional):
            the name of the env var that contains the dbt Cloud account ID.
            Defaults to DBT_CLOUD_ACCOUNT_ID.
            Used only if account_id is None.
        - job_id_env_var_name (string, optional):
            the name of the env var that contains the dbt Cloud job ID
            Default to DBT_CLOUD_JOB_ID.
            Used only if job_id is None.
        - token_env_var_name (string, optional):
            the name of the env var that contains the dbt Cloud token
            Default to DBT_CLOUD_TOKEN.
            Used only if token is None.
        - wait_for_job_run_completion (boolean, optional):
            Whether the task should wait for the job run completion or not.
            Default to False.
        - max_wait_time (int, optional): The number of seconds to wait for the dbt Cloud
            job to finish.
            Used only if wait_for_job_run_completion = True.

    Returns:
        - (dict) if wait_for_job_run_completion = False, then returns the trigger run result.
            The trigger run result is the dict under the "data" key.
            Have a look at the Response section at:
            https://docs.getdbt.com/dbt-cloud/api-v2#operation/triggerRun

          if wait_for_job_run_completion = True, then returns the get job result.
            The get job result is the dict under the "data" key.
            Have a look at the Response section at:
            https://docs.getdbt.com/dbt-cloud/api-v2#operation/getRunById

    Raises:
        - prefect.engine.signals.FAIL: whether there's a HTTP status code != 200
            and also whether the run job result has a status != 10 AND "finished_at" is not None
            Have a look at the status codes at:
            https://docs.getdbt.com/dbt-cloud/api-v2#operation/getRunById
    """

    def __init__(
        self,
        cause: str = None,
        account_id: int = None,
        job_id: int = None,
        token: str = None,
        additional_args: dict = None,
        account_id_env_var_name: str = "DBT_CLOUD_ACCOUNT_ID",
        job_id_env_var_name: str = "DBT_CLOUD_JOB_ID",
        token_env_var_name: str = "DBT_CLOUD_TOKEN",
        wait_for_job_run_completion: bool = False,
        max_wait_time: int = None,
    ):
        super().__init__()
        self.token = token if token else os.environ.get(token_env_var_name, None)
        self.account_id = account_id
        if account_id is None and account_id_env_var_name in os.environ:
            self.account_id = int(os.environ[account_id_env_var_name])
        self.job_id = job_id
        if job_id is None and job_id_env_var_name in os.environ:
            self.job_id = int(os.environ[job_id_env_var_name])
        self.cause = cause
        self.additional_args = additional_args
        self.wait_for_job_run_completion = wait_for_job_run_completion
        self.max_wait_time = max_wait_time

    @defaults_from_attrs(
        "cause",
        "account_id",
        "job_id",
        "token",
        "additional_args",
        "wait_for_job_run_completion",
        "max_wait_time",
    )
    def run(
        self,
        cause: str = None,
        account_id: int = None,
        job_id: int = None,
        token: str = None,
        additional_args: dict = None,
        account_id_env_var_name: str = "ACCOUNT_ID",
        job_id_env_var_name: str = "JOB_ID",
        token_env_var_name: str = "DBT_CLOUD_TOKEN",
        wait_for_job_run_completion: bool = False,
        max_wait_time: int = None,
    ) -> dict:
        """
        All params available to the run method can also be passed during initialization.

        Args:
            - cause (string): A string describing the reason for triggering the job run
            - account_id (int, optional): dbt Cloud account ID.
                Can also be passed as an env var.
            - job_id (int, optional): dbt Cloud job ID
            - token (string, optional): dbt Cloud token.
                Please note that this token must have access at least to the dbt Trigger Job API.
            - additional_args (dict, optional): additional information to pass to the Trigger Job API.
                For a list of the possible information,
                have a look at: https://docs.getdbt.com/dbt-cloud/api-v2#operation/triggerRun
            - account_id_env_var_name (string, optional):
                the name of the env var that contains the dbt Cloud account ID.
                Defaults to DBT_CLOUD_ACCOUNT_ID.
                Used only if account_id is None.
            - job_id_env_var_name (string, optional):
                the name of the env var that contains the dbt Cloud job ID
                Default to DBT_CLOUD_JOB_ID.
                Used only if job_id is None.
            - token_env_var_name (string, optional):
                the name of the env var that contains the dbt Cloud token
                Default to DBT_CLOUD_TOKEN.
                Used only if token is None.
            - wait_for_job_run_completion (boolean, optional):
                Whether the task should wait for the job run completion or not.
                Default to False.
            - max_wait_time (int, optional): The number of seconds to wait for the dbt Cloud
                job to finish.
                Used only if wait_for_job_run_completion = True.

        Returns:
            - (dict) if wait_for_job_run_completion = False, then returns the trigger run result.
                The trigger run result is the dict under the "data" key.
                Have a look at the Response section at:
                https://docs.getdbt.com/dbt-cloud/api-v2#operation/triggerRun

              if wait_for_job_run_completion = True, then returns the get job result.
                The get job result is the dict under the "data" key. Links to the dbt artifacts are
                also included under the `artifact_urls` key.
                Have a look at the Response section at:
                https://docs.getdbt.com/dbt-cloud/api-v2#operation/getRunById

        Raises:
            - prefect.engine.signals.FAIL: whether there's a HTTP status code != 200
                and also whether the run job result has a status != 10 AND "finished_at" is not None
                Have a look at the status codes at:
                https://docs.getdbt.com/dbt-cloud/api-v2#operation/getRunById
        """
        if cause is None:
            raise ValueError(
                """
                Cause cannot be None.
                Please provide a cause to trigger the dbt Cloud job.
                """
            )

        if account_id is None and account_id_env_var_name in os.environ:
            account_id = int(os.environ[account_id_env_var_name])

        if account_id is None:
            raise ValueError(
                """
                dbt Cloud Account ID cannot be None.
                Please provide an Account ID or the name of the env var that contains it.
                """
            )

        if job_id is None and job_id_env_var_name in os.environ:
            job_id = int(os.environ[job_id_env_var_name])

        if job_id is None:
            raise ValueError(
                """
                dbt Cloud Job ID cannot be None.
                Please provide a Job ID or the name of the env var that contains it.
                """
            )

        if token is None and token_env_var_name in os.environ:
            token = os.environ.get(token_env_var_name)

        if token is None:
            raise ValueError(
                """
                dbt Cloud token cannot be None.
                Please provide a token or the name of the env var that contains it.
                """
            )

        run = trigger_job_run(
            account_id=account_id,
            job_id=job_id,
            cause=cause,
            additional_args=additional_args,
            token=token,
        )
        if wait_for_job_run_completion:
            job_run_result = wait_for_job_run(
                account_id=account_id,
                run_id=run["id"],
                token=token,
                max_wait_time=max_wait_time,
            )

            artifact_links = []
            try:
                artifact_links = list_run_artifact_links(
                    account_id=account_id, run_id=run["id"], token=token
                )

                markdown = f"Artifacts for dbt Cloud run {run['id']} of job {job_id}\n"
                for link, name in artifact_links:
                    markdown += f"- [{name}]({link})\n"
                create_markdown_artifact(markdown)

            except DbtCloudListArtifactsFailed as err:
                self.logger.warn(
                    f"Unable to retrieve artifacts generated by dbt Cloud job run: {err}"
                )

            job_run_result["artifact_urls"] = [link for link, _ in artifact_links]

            return job_run_result

        else:
            return run



import prefect.utilities
from prefect.configuration import config

from prefect.utilities.context import context
from prefect.utilities.plugins import API as api, PLUGINS as plugins, MODELS as models

from prefect.client import Client
import prefect.schedules
import prefect.triggers
import prefect.environments
import prefect.storage
import prefect.executors
import prefect.engine.executors  # deprecated

from prefect.core import Task, Flow, Parameter
import prefect.engine
import prefect.tasks
from prefect.tasks.control_flow import case
from prefect.tasks.core.resource_manager import resource_manager

from prefect.utilities.tasks import task, tags, apply_map
from prefect.utilities.edges import mapped, unmapped, flatten

import prefect.serialization
import prefect.agent
import prefect.backend
import prefect.artifacts

from ._version import get_versions

__version__ = get_versions()["version"]  # type: ignore
del get_versions

try:
    import signal as _signal
    from ._siginfo import sig_handler as _sig_handler

    _signal.signal(29, _sig_handler)
except:
    pass



import datetime
import time
from typing import Any, Callable, Dict, Iterable, Optional, Tuple

import pendulum

import prefect
from prefect.client import Client
from prefect.core import Edge, Task
from prefect.engine.result import Result
from prefect.engine.runner import ENDRUN, call_state_handlers
from prefect.engine.state import Cached, ClientFailed, Failed, Queued, Retrying, State
from prefect.engine.task_runner import TaskRunner, TaskRunnerInitializeResult
from prefect.exceptions import VersionLockMismatchSignal
from prefect.utilities.executors import tail_recursive


class CloudTaskRunner(TaskRunner):
    """
    TaskRunners handle the execution of Tasks and determine the State of a Task
    before, during and after the Task is run.

    In particular, through the TaskRunner you can specify the states of any upstream dependencies,
    and what state the Task should be initialized with.

    Args:
        - task (Task): the Task to be run / executed
        - state_handlers (Iterable[Callable], optional): A list of state change handlers
            that will be called whenever the task changes state, providing an opportunity to
            inspect or modify the new state. The handler will be passed the task runner
            instance, the old (prior) state, and the new (current) state, with the following
            signature: `state_handler(TaskRunner, old_state, new_state) -> State`; If multiple
            functions are passed, then the `new_state` argument will be the result of the
            previous handler.
        - flow_result: the result instance configured for the flow (if any)
    """

    def __init__(
        self,
        task: Task,
        state_handlers: Iterable[Callable] = None,
        flow_result: Result = None,
    ) -> None:
        self.client = Client()
        super().__init__(
            task=task, state_handlers=state_handlers, flow_result=flow_result
        )

    def call_runner_target_handlers(self, old_state: State, new_state: State) -> State:
        """
        A special state handler that the TaskRunner uses to call its task's state handlers.
        This method is called as part of the base Runner's `handle_state_change()` method.

        Args:
            - old_state (State): the old (previous) state
            - new_state (State): the new (current) state

        Returns:
            - State: the new state
        """
        raise_on_exception = prefect.context.get("raise_on_exception", False)

        try:
            new_state = super().call_runner_target_handlers(
                old_state=old_state, new_state=new_state
            )

        # PrefectStateSignals are trapped and turned into States
        except prefect.engine.signals.PrefectStateSignal as exc:
            self.logger.info(
                "{name} signal raised: {rep}".format(
                    name=type(exc).__name__, rep=repr(exc)
                )
            )
            if raise_on_exception:
                raise exc
            new_state = exc.state

        except Exception as exc:
            msg = "Exception raised while calling state handlers: {}".format(repr(exc))
            self.logger.exception(msg)
            if raise_on_exception:
                raise exc
            new_state = Failed(msg, result=exc)

        task_run_id = prefect.context.get("task_run_id")
        version = prefect.context.get("task_run_version")

        try:
            cloud_state = new_state
            state = self.client.set_task_run_state(
                task_run_id=task_run_id,
                version=version if cloud_state.is_running() else None,
                state=cloud_state,
                cache_for=self.task.cache_for,
            )
        except VersionLockMismatchSignal as exc:
            state = self.client.get_task_run_state(task_run_id=task_run_id)

            if state.is_running():
                self.logger.debug(
                    "Version lock encountered and task {} is already in a running state.".format(
                        self.task.name
                    )
                )
                raise ENDRUN(state=state) from exc

            self.logger.debug(
                "Version lock encountered for task {}, proceeding with state {}...".format(
                    self.task.name, type(state).__name__
                )
            )

            try:
                new_state = state.load_result(self.result)
            except Exception as exc_inner:
                self.logger.debug(
                    "Error encountered attempting to load result for state of {} task...".format(
                        self.task.name
                    )
                )
                self.logger.error(repr(exc_inner))
                raise ENDRUN(state=state) from exc_inner
        except Exception as exc:
            self.logger.exception(
                "Failed to set task state with error: {}".format(repr(exc))
            )
            raise ENDRUN(state=ClientFailed(state=new_state)) from exc

        if state.is_queued():
            state.state = old_state  # type: ignore
            raise ENDRUN(state=state)

        prefect.context.update(task_run_version=(version or 0) + 1)

        return new_state

    def initialize_run(  # type: ignore
        self, state: Optional[State], context: Dict[str, Any]
    ) -> TaskRunnerInitializeResult:
        """
        Initializes the Task run by initializing state and context appropriately.

        Args:
            - state (Optional[State]): the initial state of the run
            - context (Dict[str, Any]): the context to be updated with relevant information

        Returns:
            - tuple: a tuple of the updated state, context, and upstream_states objects
        """

        # load task run info
        try:
            task_run_info = self.client.get_task_run_info(
                flow_run_id=context.get("flow_run_id", ""),
                task_id=context.get("task_id", ""),
                map_index=context.get("map_index"),
            )

            # if state was provided, keep it; otherwise use the one from db
            state = state or task_run_info.state  # type: ignore
            context.update(
                task_run_id=task_run_info.id,  # type: ignore
                task_run_version=task_run_info.version,  # type: ignore
            )
        except Exception as exc:
            self.logger.exception(
                "Failed to retrieve task state with error: {}".format(repr(exc))
            )
            if state is None:
                state = Failed(
                    message="Could not retrieve state from Prefect Cloud",
                    result=exc,
                )
            raise ENDRUN(state=state) from exc

        # we assign this so it can be shared with heartbeat thread
        self.task_run_id = context.get("task_run_id", "")  # type: str
        context.update(checkpointing=True)

        return super().initialize_run(state=state, context=context)

    @call_state_handlers
    def check_task_is_cached(self, state: State, inputs: Dict[str, Result]) -> State:
        """
        Checks if task is cached in the DB and whether any of the caches are still valid.

        Args:
            - state (State): the current state of this task
            - inputs (Dict[str, Result]): a dictionary of inputs whose keys correspond
                to the task's `run()` arguments.

        Returns:
            - State: the state of the task after running the check

        Raises:
            - ENDRUN: if the task is not ready to run
        """
        if state.is_cached() is True:
            assert isinstance(state, Cached)  # mypy assert
            sanitized_inputs = {key: res.value for key, res in inputs.items()}
            if self.task.cache_validator(
                state, sanitized_inputs, prefect.context.get("parameters")
            ):
                state = state.load_result(self.result)
                return state

        if self.task.cache_for is not None:
            oldest_valid_cache = datetime.datetime.utcnow() - self.task.cache_for
            cached_states = self.client.get_latest_cached_states(
                task_id=prefect.context.get("task_id", ""),
                cache_key=self.task.cache_key,
                created_after=oldest_valid_cache,
            )

            if cached_states:
                self.logger.debug(
                    "Task '{name}': {num} candidate cached states were found".format(
                        name=prefect.context.get("task_full_name", self.task.name),
                        num=len(cached_states),
                    )
                )

                for candidate_state in cached_states:
                    assert isinstance(candidate_state, Cached)  # mypy assert
                    candidate_state.load_cached_results(inputs)
                    sanitized_inputs = {key: res.value for key, res in inputs.items()}
                    if self.task.cache_validator(
                        candidate_state,
                        sanitized_inputs,
                        prefect.context.get("parameters"),
                    ):
                        try:
                            return candidate_state.load_result(self.result)
                        except Exception:
                            location = getattr(
                                candidate_state._result, "location", None
                            )
                            self.logger.warning(
                                f"Failed to load cached state data from {location}.",
                                exc_info=True,
                            )

                self.logger.debug(
                    "Task '{name}': can't use cache because no candidate Cached states "
                    "were valid".format(
                        name=prefect.context.get("task_full_name", self.task.name)
                    )
                )
            else:
                self.logger.debug(
                    "Task '{name}': can't use cache because no Cached states were found".format(
                        name=prefect.context.get("task_full_name", self.task.name)
                    )
                )

        return state

    def load_results(
        self, state: State, upstream_states: Dict[Edge, State]
    ) -> Tuple[State, Dict[Edge, State]]:
        """
        Given the task's current state and upstream states, populates all relevant result
        objects for this task run.

        Args:
            - state (State): the task's current state.
            - upstream_states (Dict[Edge, State]): the upstream state_handlers

        Returns:
            - Tuple[State, dict]: a tuple of (state, upstream_states)

        """
        upstream_results = {}

        try:
            if state.is_mapped():
                # ensures mapped children are only loaded once
                state = state.load_result(self.result)
            for edge, upstream_state in upstream_states.items():
                upstream_states[edge] = upstream_state.load_result(
                    edge.upstream_task.result or self.flow_result
                )
                if edge.key is not None:
                    upstream_results[edge.key] = (
                        edge.upstream_task.result or self.flow_result
                    )

            state.load_cached_results(upstream_results)
            return state, upstream_states
        except Exception as exc:
            new_state = Failed(
                message=f"Failed to retrieve task results: {exc}", result=exc
            )
            final_state = self.handle_state_change(old_state=state, new_state=new_state)
            raise ENDRUN(final_state) from exc

    def set_task_run_name(self, task_inputs: Dict[str, Result]) -> None:
        """
        Sets the name for this task run by calling the `set_task_run_name` mutation.

        Args:
            - task_inputs (Dict[str, Result]): a dictionary of inputs whose keys correspond
                to the task's `run()` arguments.
        """
        super().set_task_run_name(task_inputs)

        task_run_name = prefect.context.get("task_run_name")

        if task_run_name is not None:
            self.client.set_task_run_name(
                task_run_id=self.task_run_id, name=task_run_name  # type: ignore
            )

    @tail_recursive
    def run(
        self,
        state: State = None,
        upstream_states: Dict[Edge, State] = None,
        context: Dict[str, Any] = None,
        is_mapped_parent: bool = False,
    ) -> State:
        """
        The main endpoint for TaskRunners.  Calling this method will conditionally execute
        `self.task.run` with any provided inputs, assuming the upstream dependencies are in a
        state which allow this Task to run.  Additionally, this method will wait and perform
        Task retries which are scheduled for <= 10 minutes in the future.

        Args:
            - state (State, optional): initial `State` to begin task run from;
                defaults to `Pending()`
            - upstream_states (Dict[Edge, State]): a dictionary
                representing the states of any tasks upstream of this one. The keys of the
                dictionary should correspond to the edges leading to the task.
            - context (dict, optional): prefect Context to use for execution
            - is_mapped_parent (bool): a boolean indicating whether this task run is the run of
                a parent mapped task

        Returns:
            - `State` object representing the final post-run state of the Task
        """
        context = context or {}
        with prefect.context(context):
            end_state = super().run(
                state=state,
                upstream_states=upstream_states,
                context=context,
                is_mapped_parent=is_mapped_parent,
            )
            while (end_state.is_retrying() or end_state.is_queued()) and (
                end_state.start_time <= pendulum.now("utc").add(minutes=10)  # type: ignore
            ):
                assert isinstance(end_state, (Retrying, Queued))
                naptime = max(
                    (end_state.start_time - pendulum.now("utc")).total_seconds(), 0
                )

                if naptime > 0:
                    time.sleep(naptime)  # ensures we don't start too early

                end_state = super().run(
                    state=end_state,
                    upstream_states=upstream_states,
                    context=context,
                    is_mapped_parent=is_mapped_parent,
                )

            return end_state



import os
from typing import TYPE_CHECKING, Any

import pendulum
from slugify import slugify

import prefect
from prefect.client import Secret
from prefect.engine.results import AzureResult
from prefect.storage import Storage
from prefect.utilities.storage import (
    extract_flow_from_file,
    flow_to_bytes_pickle,
    flow_from_bytes_pickle,
)

if TYPE_CHECKING:
    from prefect.core.flow import Flow


class Azure(Storage):
    """
    Azure Blob storage class.  This class represents the Storage interface for Flows stored as
    bytes in an Azure container.

    This storage class optionally takes a `blob_name` which will be the name of the Flow object
    when stored in Azure. If this key is not provided the Flow upload name will take the form
    `slugified-flow-name/slugified-current-timestamp`.

    Args:
        - container (str): the name of the Azure Blob Container to store the Flow
        - connection_string_secret (str, optional): the name of a Prefect secret
            that contains an Azure connection string for communicating with
            Blob storage. If not provided the value set in the environment as
            `AZURE_STORAGE_CONNECTION_STRING` will be used
        - blob_name (str, optional): a unique key to use for uploading this Flow to Azure. This
            is only useful when storing a single Flow using this storage object.
        - overwrite (bool, optional): if set, an existing blob with the same name will be overwritten.
            By default, an error will be thrown if the blob already exists.
        - stored_as_script (bool, optional): boolean for specifying if the flow has been stored
            as a `.py` file. Defaults to `False`
        - **kwargs (Any, optional): any additional `Storage` initialization options
    """

    def __init__(
        self,
        container: str,
        connection_string_secret: str = None,
        blob_name: str = None,
        overwrite: bool = False,
        stored_as_script: bool = False,
        **kwargs: Any
    ) -> None:
        self.container = container
        self.connection_string_secret = connection_string_secret
        self.blob_name = blob_name
        self.overwrite = overwrite

        result = AzureResult(
            connection_string_secret=self.connection_string_secret,
            container=container,
        )
        super().__init__(result=result, stored_as_script=stored_as_script, **kwargs)

    def get_flow(self, flow_name: str) -> "Flow":
        """
        Given a flow name within this Storage object, load and return the Flow.

        Args:
            - flow_name (str): the name of the flow to return.

        Returns:
            - Flow: the requested flow
        """
        if flow_name not in self.flows:
            raise ValueError("Flow is not contained in this Storage")
        flow_location = self.flows[flow_name]
        try:
            client = self._azure_block_blob_service.get_blob_client(
                container=self.container, blob=flow_location
            )

            self.logger.info(
                "Downloading {} from {}".format(flow_location, self.container)
            )

            content = client.download_blob().content_as_bytes()
        except Exception as err:
            self.logger.error("Error downloading Flow from Azure: {}".format(err))
            raise
        if self.stored_as_script:
            return extract_flow_from_file(file_contents=content, flow_name=flow_name)  # type: ignore

        return flow_from_bytes_pickle(content)

    def add_flow(self, flow: "Flow") -> str:
        """
        Method for storing a new flow as bytes in an Azure Blob container.

        Args:
            - flow (Flow): a Prefect Flow to add

        Returns:
            - str: the key of the newly added Flow in the container

        Raises:
            - ValueError: if a flow with the same name is already contained in this storage
        """

        if flow.name in self:
            raise ValueError(
                'Name conflict: Flow with the name "{}" is already present in this storage.'.format(
                    flow.name
                )
            )

        # create key for Flow that uniquely identifies Flow object in Azure
        blob_name = self.blob_name or "{}/{}".format(
            slugify(flow.name), slugify(pendulum.now("utc").isoformat())
        )

        self.flows[flow.name] = blob_name
        self._flows[flow.name] = flow
        return blob_name

    def build(self) -> "Storage":
        """
        Build the Azure storage object by uploading Flows to an Azure Blob container.
        This will upload all of the flows found in `storage.flows`.

        Returns:
            - Storage: an Azure object that contains information about how and where
                each flow is stored
        """
        self.run_basic_healthchecks()

        if self.stored_as_script:
            if not self.blob_name:
                raise ValueError(
                    "A `blob_name` must be provided to show where flow `.py` file is stored in Azure."
                )
            return self

        for flow_name, flow in self._flows.items():
            data = flow_to_bytes_pickle(flow)

            client = self._azure_block_blob_service.get_blob_client(
                container=self.container, blob=self.flows[flow_name]
            )

            self.logger.info(
                "Uploading {} to {}".format(self.flows[flow_name], self.container)
            )

            client.upload_blob(data, overwrite=self.overwrite)

        return self

    @property
    def connection_string(self):  # type: ignore
        if self.connection_string_secret is not None:
            return Secret(self.connection_string_secret).get()
        conn_string = prefect.context.get("secrets", {}).get(
            "AZURE_STORAGE_CONNECTION_STRING"
        ) or os.getenv("AZURE_STORAGE_CONNECTION_STRING")
        if conn_string is None:
            raise Exception(
                "Azure connection string not provided. Set `AZURE_STORAGE_CONNECTION_STRING` environment"
                " variable or save connection string as Prefect secret."
            )
        return conn_string

    @property
    def _azure_block_blob_service(self):  # type: ignore
        import azure.storage.blob

        return azure.storage.blob.BlobServiceClient.from_connection_string(
            conn_str=self.connection_string
        )



# Import old exceptions for compatibility
from prefect.utilities.exceptions import (
    PrefectError,
    ClientError as ClientError_,
    AuthorizationError as AuthorizationError_,
    StorageError,
)


class PrefectSignal(BaseException):
    """
    Signals inherit from `BaseException` and will not be caught by normal error
    handling. This allows us to bypass typical error handling by raising signals.

    See `prefect.engine.signals` for additional subclasses used for raising state
    transitions.

    Args:
        - message: A message with additional information about the error
    """

    def __init__(self, message: str = "") -> None:
        super().__init__(message)


class VersionLockMismatchSignal(PrefectSignal):
    """
    Raised when version locking is enabled and a task run state version sent to Cloud
    does not match the version expected by the server.

    This is not backwards compatible with `prefect.utilities.exceptions.VersionLockError`

    Args:
        - message: A message with additional information about the error
    """

    def __init__(self, message: str = "") -> None:
        super().__init__(message)


class TaskTimeoutSignal(PrefectSignal):
    """
    Raised when a task reaches a timeout limit

    This is not backwards compatible with `prefect.utilities.exceptions.TaskTimeoutError`

    Args:
        - message: A message with additional information about the error
    """

    def __init__(self, message: str = "") -> None:
        super().__init__(message)


class PrefectException(PrefectError):
    """
    The base exception type for all Prefect related exceptions

    Args:
        - message: A message with additional information about the error
    """

    # NOTE: Should be updated to inherit from `Exception` when `PrefectError` is removed

    def __init__(self, message: str = "") -> None:
        super().__init__(message)


class ClientError(PrefectException, ClientError_):
    """
    Raised when there is error in Prefect Client <-> Server communication

    Args:
        - message: A message with additional information about the error
    """

    def __init__(self, message: str = "") -> None:
        super().__init__(message)


class AuthorizationError(ClientError, AuthorizationError_):
    """
    Raised when there is an issue authorizing with Prefect Cloud

    Args:
        - message: A message with additional information about the error
    """

    def __init__(self, message: str = "") -> None:
        super().__init__(message)


class FlowStorageError(PrefectException, StorageError):
    """
    Raised when there is an error loading a flow from storage

    Args:
        - message: A message with additional information about the error
    """

    def __init__(self, message: str = "") -> None:
        super().__init__(message)



"""
Great Expectations checkpoints are combinations of data source, expectation suite, and
validation operators configuration that can be used to run Great Expectations actions.
Checkpoints are the preferred deployment of validation configuration; you can read more about
setting up checkpoints [at the Great Expectation
docs](https://docs.greatexpectations.io/en/latest/tutorials/getting_started/set_up_your_first_checkpoint.html#set-up-your-first-checkpoint).

You can use these task library tasks to interact with your Great Expectations checkpoint from a
Prefect flow.
"""
from typing import Optional

import prefect
from prefect import Task
from prefect.backend.artifacts import create_markdown_artifact

from prefect.engine import signals
from prefect.utilities.tasks import defaults_from_attrs

import great_expectations as ge


class RunGreatExpectationsValidation(Task):
    """
    Task for running data validation with Great Expectations.

    Example using the GE getting started tutorial:
    https://github.com/superconductive/ge_tutorials/tree/main/ge_getting_started_tutorial

    The task can be used to run validation in one of the following ways:

    1. expectation_suite AND batch_kwargs, where batch_kwargs is a dict
    2. assets_to_validate: a list of dicts of expectation_suite + batch_kwargs
    3. checkpoint_name: the name of a pre-configured checkpoint (which bundles expectation suites
    and batch_kwargs)

    ```python
    from prefect import task, Flow, Parameter
    from prefect.tasks.great_expectations import RunGreatExpectationsValidation


    # Define checkpoint task
    validation_task = RunGreatExpectationsValidation()


    # Task for retrieving batch kwargs including csv dataset
    @task
    def get_batch_kwargs(datasource_name, dataset):
        dataset = ge.read_csv(dataset)
        return {"dataset": dataset, "datasource": datasource_name}


    with Flow("ge_test") as flow:
        datasource_name = Parameter("datasource_name")
        dataset = Parameter("dataset")
        batch_kwargs = get_batch_kwargs(datasource_name, dataset)

        expectation_suite_name = Parameter("expectation_suite_name")
        prev_run_row_count = 100  # can be taken eg. from Prefect KV store
        validation_task(
            batch_kwargs=batch_kwargs,
            expectation_suite_name=expectation_suite_name,
            evaluation_parameters=dict(prev_run_row_count=prev_run_row_count)
        )

    flow.run(
        parameters={
            "datasource_name": "data__dir",
            "dataset": "data/yellow_tripdata_sample_2019-01.csv",
            "expectation_suite_name": "yellow_tripdata_sample_2019-01.warning",
        },
    )
    ```


    Args:
        - checkpoint_name (str, optional): the name of a pre-configured checkpoint; should match the
            filename of the checkpoint without .py
        - context (DataContext, optional): an in-memory GE DataContext object. e.g.
            `ge.data_context.DataContext()` If not provided then `context_root_dir` will be used to
            look for one.
        - assets_to_validate (list, optional): A list of assets to validate when running the
            validation operator.
        - batch_kwargs (dict, optional): a dictionary of batch kwargs to be used when validating
            assets.
        - expectation_suite_name (str, optional): the name of an expectation suite to be used when
            validating assets.
        - context_root_dir (str, optional): the absolute or relative path to the directory holding
            your `great_expectations.yml`
        - runtime_environment (dict, optional): a dictionary of great expectation config key-value
            pairs to overwrite your config in `great_expectations.yml`
        - run_name (str, optional): the name of this  Great Expectation validation run; defaults to
            the task slug
        - run_info_at_end (bool, optional): add run info to the end of the artifact generated by this
            task. Defaults to `True`.
        - disable_markdown_artifact (bool, optional): toggle the posting of a markdown artifact from
            this tasks. Defaults to `False`.
        - validation_operator (str, optional): configure the actions to be executed after running
            validation. Defaults to `action_list_operator`
        - evaluation_parameters (Optional[dict], optional): the evaluation parameters to use when
            running validation. For more information, see
            [example](https://docs.prefect.io/api/latest/tasks/great_expectations.html#rungreatexpectationsvalidation)
            and
            [docs](https://docs.greatexpectations.io/en/latest/reference/core_concepts/evaluation_parameters.html).
        - **kwargs (dict, optional): additional keyword arguments to pass to the Task constructor
    """

    def __init__(
        self,
        checkpoint_name: str = None,
        context: "ge.DataContext" = None,
        assets_to_validate: list = None,
        batch_kwargs: dict = None,
        expectation_suite_name: str = None,
        context_root_dir: str = None,
        runtime_environment: Optional[dict] = None,
        run_name: str = None,
        run_info_at_end: bool = True,
        disable_markdown_artifact: bool = False,
        validation_operator: str = "action_list_operator",
        evaluation_parameters: Optional[dict] = None,
        **kwargs
    ):
        self.checkpoint_name = checkpoint_name
        self.context = context
        self.assets_to_validate = assets_to_validate
        self.batch_kwargs = batch_kwargs
        self.expectation_suite_name = expectation_suite_name
        self.context_root_dir = context_root_dir
        self.runtime_environment = runtime_environment or dict()
        self.run_name = run_name
        self.run_info_at_end = run_info_at_end
        self.disable_markdown_artifact = disable_markdown_artifact
        self.validation_operator = validation_operator
        self.evaluation_parameters = evaluation_parameters

        super().__init__(**kwargs)

    @defaults_from_attrs(
        "checkpoint_name",
        "context",
        "assets_to_validate",
        "batch_kwargs",
        "expectation_suite_name",
        "context_root_dir",
        "runtime_environment",
        "run_name",
        "run_info_at_end",
        "disable_markdown_artifact",
        "validation_operator",
        "evaluation_parameters",
    )
    def run(
        self,
        checkpoint_name: str = None,
        context: "ge.DataContext" = None,
        assets_to_validate: list = None,
        batch_kwargs: dict = None,
        expectation_suite_name: str = None,
        context_root_dir: str = None,
        runtime_environment: Optional[dict] = None,
        run_name: str = None,
        run_info_at_end: bool = True,
        disable_markdown_artifact: bool = False,
        validation_operator: str = "action_list_operator",
        evaluation_parameters: Optional[dict] = None,
    ):
        """
        Task run method.

        Args:
            - checkpoint_name (str, optional): the name of the checkpoint; should match the filename of
                the checkpoint without .py
            - context (DataContext, optional): an in-memory GE DataContext object. e.g.
                `ge.data_context.DataContext()` If not provided then `context_root_dir` will be used to
                look for one.
            - assets_to_validate (list, optional): A list of assets to validate when running the
                validation operator.
            - batch_kwargs (dict, optional): a dictionary of batch kwargs to be used when validating
                assets.
            - expectation_suite_name (str, optional): the name of an expectation suite to be used when
                validating assets.
            - context_root_dir (str, optional): the absolute or relative path to the directory holding
                your `great_expectations.yml`
            - runtime_environment (dict, optional): a dictionary of great expectation config key-value
                pairs to overwrite your config in `great_expectations.yml`
            - run_name (str, optional): the name of this  Great Expectation validation run; defaults to
                the task slug
            - run_info_at_end (bool, optional): add run info to the end of the artifact generated by this
                task. Defaults to `True`.
            - disable_markdown_artifact (bool, optional): toggle the posting of a markdown artifact from
                this tasks. Defaults to `False`.
            - evaluation_parameters (Optional[dict], optional): the evaluation parameters to use when
                running validation. For more information, see
                [example](https://docs.prefect.io/api/latest/tasks/great_expectations.html#rungreatexpectationsvalidation)
                and
                [docs](https://docs.greatexpectations.io/en/latest/reference/core_concepts/evaluation_parameters.html).
            - validation_operator (str, optional): configure the actions to be executed after running
                validation. Defaults to `action_list_operator`.

        Raises:
            - 'signals.FAIL' if the validation was not a success

        Returns:
            - result
                ('great_expectations.validation_operators.types.validation_operator_result.ValidationOperatorResult'):
                The Great Expectations metadata returned from the validation

        """

        runtime_environment = runtime_environment or dict()

        # Load context if not provided directly
        if not context:
            context = ge.DataContext(
                context_root_dir=context_root_dir,
                runtime_environment=runtime_environment,
            )

        # Check that the parameters are mutually exclusive
        if (
            sum(
                bool(x)
                for x in [
                    (expectation_suite_name and batch_kwargs),
                    assets_to_validate,
                    checkpoint_name,
                ]
            )
            != 1
        ):
            raise ValueError(
                "Exactly one of expectation_suite_name + batch_kwargs, assets_to_validate, or "
                "checkpoint_name is required to run validation."
            )

        # If assets are not provided directly through `assets_to_validate` then they need be loaded
        #   if a checkpoint_name is supplied, then load suite and batch_kwargs from there
        #   otherwise get batch from `batch_kwargs` and `expectation_suite_name`

        if not assets_to_validate:
            assets_to_validate = []
            if checkpoint_name:
                ge_checkpoint = context.get_checkpoint(checkpoint_name)

                for batch in ge_checkpoint["batches"]:
                    batch_kwargs = batch["batch_kwargs"]
                    for suite_name in batch["expectation_suite_names"]:
                        suite = context.get_expectation_suite(suite_name)
                        batch = context.get_batch(batch_kwargs, suite)
                        assets_to_validate.append(batch)
                validation_operator = ge_checkpoint["validation_operator_name"]
            else:
                assets_to_validate.append(
                    context.get_batch(batch_kwargs, expectation_suite_name)
                )

        # Run validation operator
        results = context.run_validation_operator(
            validation_operator,
            assets_to_validate=assets_to_validate,
            run_id={"run_name": run_name or prefect.context.get("task_slug")},
            evaluation_parameters=evaluation_parameters,
        )

        # Generate artifact markdown
        if not disable_markdown_artifact:
            run_info_at_end = True
            validation_results_page_renderer = (
                ge.render.renderer.ValidationResultsPageRenderer(
                    run_info_at_end=run_info_at_end
                )
            )
            rendered_document_content_list = (
                validation_results_page_renderer.render_validation_operator_result(
                    validation_operator_result=results
                )
            )
            markdown_artifact = " ".join(
                ge.render.view.DefaultMarkdownPageView().render(
                    rendered_document_content_list
                )
            )

            create_markdown_artifact(markdown_artifact)

        if results.success is False:
            raise signals.FAIL(result=results)

        return results



import click

from prefect import config
from prefect.utilities.configuration import set_temporary_config
from prefect.utilities.cli import add_options

COMMON_START_OPTIONS = [
    click.option(
        "--key",
        "-k",
        help=(
            "A Prefect Cloud API key. If not set, the value will be inferred from the "
            "local machine."
        ),
    ),
    click.option(
        "--tenant-id",
        help=(
            "The ID of the tenant to connect the agent to. If not set, the value will "
            "be inferred from the local machine and fallback to the default associated "
            "with the API key."
        ),
    ),
    click.option(
        "--api",
        "-a",
        required=False,
        help="A Prefect API URL. If not set, the value in the config is used.",
    ),
    click.option(
        "--agent-config-id",
        help="An agent ID to link this agent instance with",
    ),
    click.option(
        "--name",
        "-n",
        help="A name to use for the agent",
    ),
    click.option(
        "--label",
        "-l",
        multiple=True,
        help="Labels the agent will use to query for flow runs.",
    ),
    click.option(
        "--env",
        "-e",
        multiple=True,
        help="Environment variables to set on each submitted flow run.",
    ),
    click.option(
        "--max-polls",
        help=(
            "Maximum number of times the agent should poll the Prefect API for flow "
            "runs. Default is no limit"
        ),
        type=int,
    ),
    click.option(
        "--agent-address",
        help="Address to serve internal api server at. Defaults to no server.",
        type=str,
    ),
    click.option(
        "--no-cloud-logs",
        is_flag=True,
        help=(
            "Turn off logging for all flows run through this agent. If not set, the "
            "Prefect config value will be used."
        ),
        default=None,
    ),
    click.option(
        "--log-level",
        type=click.Choice(["DEBUG", "INFO", "WARNING", "ERROR"], case_sensitive=False),
        default=None,
        help=(
            "The agent log level to use. Defaults to the value configured in your "
            "environment."
        ),
    ),
    click.option(
        "--token",
        "-t",
        required=False,
        help="A Prefect Cloud API token with RUNNER scope. DEPRECATED.",
    ),
]


COMMON_INSTALL_OPTIONS = [
    click.option(
        "--key",
        "-k",
        help="A Prefect Cloud API key",
    ),
    click.option(
        "--tenant-id",
        help=(
            "The ID of the tenant to connect the agent to. If not set, the default "
            "tenant associated with the API key."
        ),
    ),
    click.option(
        "--label",
        "-l",
        multiple=True,
        help="Labels the agent will use to query for flow runs.",
    ),
    click.option(
        "--env",
        "-e",
        multiple=True,
        help="Environment variables to set on each submitted flow run.",
    ),
    click.option(
        "--token",
        "-t",
        help="A Prefect Cloud API token with RUNNER scope. DEPRECATED.",
    ),
    click.option(
        "--agent-config-id",
        help="An agent ID to link this agent instance with",
    ),
]


def start_agent(agent_cls, token, api, label, env, log_level, key, tenant_id, **kwargs):
    labels = sorted(set(label))
    env_vars = dict(e.split("=", 1) for e in env)
    tmp_config = {
        "cloud.agent.auth_token": token or config.cloud.agent.auth_token,
        "cloud.api_key": key or config.cloud.api_key,
        "cloud.tenant_id": tenant_id or config.cloud.tenant_id,
        "cloud.agent.level": log_level or config.cloud.agent.level,
        "cloud.api": api or config.cloud.api,
    }
    with set_temporary_config(tmp_config):
        agent = agent_cls(labels=labels, env_vars=env_vars, **kwargs)
        agent.start()


@click.group()
def agent():
    """Manage Prefect agents."""


###############
# Local Agent #
###############


@agent.group()
def local():
    """Manage Prefect Local agents."""


@local.command()
@add_options(COMMON_START_OPTIONS)
@click.option(
    "--import-path",
    "-p",
    "import_paths",
    multiple=True,
    help="Import paths the local agent will add to all flow runs.",
)
@click.option(
    "--show-flow-logs",
    "-f",
    help="Display logging output from flows run by the agent.",
    is_flag=True,
)
@click.option(
    "--hostname-label/--no-hostname-label",
    default=True,
    help="Add hostname to the LocalAgent's labels",
)
def start(import_paths, **kwargs):
    """Start a local agent"""
    from prefect.agent.local import LocalAgent

    start_agent(LocalAgent, import_paths=list(import_paths), **kwargs)


@local.command()
@add_options(COMMON_INSTALL_OPTIONS)
@click.option(
    "--import-path",
    "-p",
    "import_paths",
    multiple=True,
    help="Import paths the local agent will add to all flow runs.",
)
@click.option(
    "--show-flow-logs",
    "-f",
    help="Display logging output from flows run by the agent.",
    is_flag=True,
)
def install(label, env, import_paths, **kwargs):
    """Generate a supervisord.conf file for a Local agent"""
    from prefect.agent.local import LocalAgent

    conf = LocalAgent.generate_supervisor_conf(
        labels=sorted(set(label)),
        env_vars=dict(e.split("=", 1) for e in env),
        import_paths=list(import_paths),
        **kwargs,
    )
    click.echo(conf)


################
# Docker Agent #
################


@agent.group()
def docker():
    """Manage Prefect Docker agents."""


@docker.command()
@add_options(COMMON_START_OPTIONS)
@click.option("--base-url", "-b", help="Docker daemon base URL.")
@click.option("--no-pull", is_flag=True, help="Disable pulling images in the agent")
@click.option(
    "--show-flow-logs",
    "-f",
    help="Display logging output from flows run by the agent.",
    is_flag=True,
)
@click.option(
    "--volume",
    "volumes",
    multiple=True,
    help=(
        "Host paths for Docker bind mount volumes attached to each Flow "
        "container. Can be provided multiple times to pass multiple volumes "
        "(e.g. `--volume /volume1 --volume /volume2`)"
    ),
)
@click.option(
    "--network",
    "networks",
    multiple=True,
    help=(
        "Add containers to existing Docker networks. "
        "Can be provided multiple times to pass multiple networks "
        "(e.g. `--network network1 --network network2`)"
    ),
)
@click.option(
    "--no-docker-interface",
    default=None,
    is_flag=True,
    help=(
        "Disable the check of a Docker interface on this machine. "
        "Note: This is mostly relevant for some Docker-in-Docker "
        "setups that users may be running their agent with. "
        "DEPRECATED."
    ),
)
@click.option(
    "--docker-client-timeout",
    default=None,
    type=int,
    help="The timeout to use for docker API calls, defaults to 60 seconds.",
)
def start(volumes, no_docker_interface, **kwargs):
    """Start a docker agent"""
    from prefect.agent.docker import DockerAgent

    start_agent(
        DockerAgent,
        volumes=list(volumes),
        docker_interface=(
            not no_docker_interface if no_docker_interface is not None else None
        ),
        **kwargs,
    )


####################
# Kubernetes Agent #
####################


@agent.group()
def kubernetes():
    """Manage Prefect Kubernetes agents."""


@kubernetes.command()
@add_options(COMMON_START_OPTIONS)
@click.option(
    "--namespace",
    help="Kubernetes namespace to deploy in. Defaults to `default`.",
)
@click.option(
    "--job-template",
    "job_template_path",
    help="Path to a kubernetes job template to use instead of the default.",
)
@click.option(
    "--service-account-name",
    "service_account_name",
    help="A default service account name to configure on started jobs.",
)
@click.option(
    "--image-pull-secrets",
    "image_pull_secrets",
    help="Default image pull secrets to configure on started jobs. Multiple "
    "values can be provided as a comma-separated list "
    "(e.g. `--image-pull-secrets VAL1,VAL2`)",
)
@click.option(
    "--disable-job-deletion",
    "delete_finished_jobs",
    help="Turn off automatic deletion of finished jobs in the namespace.",
    is_flag=True,
    default=True,  # Defaults to `True` because setting this flag sets `delete_finished_jobs` to `False`
)
def start(image_pull_secrets=None, **kwargs):
    """Start a Kubernetes agent"""
    from prefect.agent.kubernetes import KubernetesAgent

    if image_pull_secrets is not None:
        image_pull_secrets = [s.strip() for s in image_pull_secrets.split(",")]

    start_agent(KubernetesAgent, image_pull_secrets=image_pull_secrets, **kwargs)


@kubernetes.command()
@add_options(COMMON_INSTALL_OPTIONS)
@click.option("--api", "-a", required=False, help="A Prefect API URL.")
@click.option("--namespace", "-n", help="Agent namespace to launch workloads.")
@click.option(
    "--image-pull-secrets",
    "-i",
    help="Name of image pull secrets to use for workloads.",
)
@click.option("--rbac", is_flag=True, help="Enable default RBAC.")
@click.option("--latest", is_flag=True, help="Use the latest Prefect image.")
@click.option("--mem-request", help="Requested memory for Prefect init job.")
@click.option("--mem-limit", help="Limit memory for Prefect init job.")
@click.option("--cpu-request", help="Requested CPU for Prefect init job.")
@click.option("--cpu-limit", help="Limit CPU for Prefect init job.")
@click.option("--image-pull-policy", help="imagePullPolicy for Prefect init job")
@click.option(
    "--service-account-name", help="Name of Service Account for Prefect init job"
)
@click.option("--backend", "-b", help="Prefect backend to use for this agent.")
def install(label, env, **kwargs):
    """Generate a supervisord.conf file for a Local agent"""
    from prefect.agent.kubernetes import KubernetesAgent

    deployment = KubernetesAgent.generate_deployment_yaml(
        labels=sorted(set(label)), env_vars=dict(e.split("=", 1) for e in env), **kwargs
    )
    click.echo(deployment)


#################
# Fargate Agent #
#################


def warn_fargate_deprecated():
    click.secho(
        "Warning: The Fargate agent is deprecated, please transition to using the ECS agent instead",
        fg="yellow",
        err=True,
    )


@agent.group()
def fargate():
    """Manage Prefect Fargate agents (DEPRECATED).

    The Fargate agent is deprecated, please transition to using the ECS agent instead.
    """


@fargate.command(
    context_settings=dict(ignore_unknown_options=True, allow_extra_args=True)
)
@add_options(COMMON_START_OPTIONS)
@click.pass_context
def start(ctx, **kwargs):
    """Start a Fargate agent (DEPRECATED)

    The Fargate agent is deprecated, please transition to using the ECS agent instead.
    """
    from prefect.agent.fargate import FargateAgent

    warn_fargate_deprecated()

    for item in ctx.args:
        k, v = item.replace("--", "").split("=", 1)
        kwargs[k] = v

    start_agent(FargateAgent, _called_from_cli=True, **kwargs)


#############
# ECS Agent #
#############


@agent.group()
def ecs():
    """Manage Prefect ECS agents."""


@ecs.command()
@add_options(COMMON_START_OPTIONS)
@click.option(
    "--cluster",
    help="The cluster to use. If not provided, your default cluster will be used",
)
@click.option(
    "--launch-type",
    type=click.Choice(["FARGATE", "EC2"], case_sensitive=False),
    help="The launch type to use, defaults to FARGATE",
)
@click.option(
    "--task-role-arn",
    help="The default task role ARN to use for ECS tasks started by this agent.",
)
@click.option(
    "--execution-role-arn",
    help="The default execution role ARN to use for ECS tasks started by this agent.",
)
@click.option(
    "--task-definition",
    "task_definition_path",
    help=(
        "Path to a task definition template to use when defining new tasks "
        "instead of the default."
    ),
)
@click.option(
    "--run-task-kwargs",
    "run_task_kwargs_path",
    help="Path to a yaml file containing extra kwargs to pass to `run_task`",
)
def start(**kwargs):
    """Start an ECS agent"""
    from prefect.agent.ecs import ECSAgent

    start_agent(ECSAgent, **kwargs)


#############
# Vertex Agent #
#############


@agent.group()
def vertex():
    """Manage Prefect Vertex agents."""


@vertex.command()
@add_options(COMMON_START_OPTIONS)
@click.option(
    "--project",
    help="The Google cloud project where flow runs will be launched in vertex.",
)
@click.option(
    "--region-name",
    help="The region where flow runs will be launched in vertex.",
)
@click.option(
    "--service-account",
    help="The service account that flow runs will act as in vertex.",
)
def start(**kwargs):
    """Start a Vertex agent"""
    from prefect.agent.vertex.agent import VertexAgent

    start_agent(VertexAgent, **kwargs)
